from typing import Optional, Sequence

import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model, trainable
from .pwc_modules import (
    conv,
    upsample2d_as,
    rescale_flow,
    initialize_msra,
    compute_cost_volume,
)
from .pwc_modules import (
    WarpingLayer,
    FeatureExtractor,
    ContextNetwork,
    FlowEstimatorDense,
    OccContextNetwork,
    OccEstimatorDense,
)
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC_Bi_Occ_upsample


class IRRPWC(BaseModel):
    pretrained_checkpoints = {
        "chairs_occ": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-chairs_occ-02066cc4.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-things-c143e848.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-sintel-6ad65777.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-kitti-74d8868f.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 0.05,
        search_range: int = 4,
        output_level: int = 4,
        num_levels: int = 7,
        num_chs: Sequence[int] = (3, 16, 32, 64, 96, 128, 196),
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(IRRPWC, self).__init__(
            output_stride=64,
            loss_fn=MultiScaleEPE_PWC_Bi_Occ_upsample(
                train_batch_size=train_batch_size, div_flow=div_flow
            ),
            **kwargs,
        )

        self.div_flow = div_flow
        self.search_range = search_range
        self.output_level = output_level
        self.num_levels = num_levels
        self.num_chs = num_chs
        self.train_batch_size = train_batch_size

        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.num_chs)
        self.warping_layer = WarpingLayer()

        self.dim_corr = (self.search_range * 2 + 1) ** 2
        self.num_ch_in_flo = self.dim_corr + 32 + 2
        self.num_ch_in_occ = self.dim_corr + 32 + 1

        self.flow_estimators = FlowEstimatorDense(self.num_ch_in_flo)
        self.context_networks = ContextNetwork(self.num_ch_in_flo + 448 + 2)
        self.occ_estimators = OccEstimatorDense(self.num_ch_in_occ)
        self.occ_context_networks = OccContextNetwork(self.num_ch_in_occ + 448 + 1)
        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)

        self.conv_1x1 = nn.ModuleList(
            [
                conv(196, 32, kernel_size=1, stride=1, dilation=1),
                conv(128, 32, kernel_size=1, stride=1, dilation=1),
                conv(96, 32, kernel_size=1, stride=1, dilation=1),
                conv(64, 32, kernel_size=1, stride=1, dilation=1),
            ]
        )

        self.conv_1x1_1 = conv(16, 3, kernel_size=1, stride=1, dilation=1)

        self.refine_flow = RefineFlow(2 + 1 + 32)
        self.refine_occ = RefineOcc(1 + 32 + 32)

        self.corr_params = {
            "pad_size": self.search_range,
            "kernel_size": 1,
            "max_disp": self.search_range,
            "stride1": 1,
            "stride2": 1,
            "corr_multiply": 1,
        }

        initialize_msra(self.modules())

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )

        x1_raw = images[:, 0]
        x2_raw = images[:, 1]
        batch_size, _, height_im, width_im = x1_raw.size()

        # on the bottom level are original images
        x1_pyramid = self.feature_pyramid_extractor(x1_raw) + [x1_raw]
        x2_pyramid = self.feature_pyramid_extractor(x2_raw) + [x2_raw]

        # outputs
        output_dict = {}
        output_dict_eval = {}
        flows = []
        occs = []

        (
            _,
            _,
            h_x1,
            w_x1,
        ) = x1_pyramid[0].size()
        flow_f = torch.zeros(batch_size, 2, h_x1, w_x1).to(
            dtype=x1_raw.dtype, device=x1_raw.device
        )
        flow_b = torch.zeros(batch_size, 2, h_x1, w_x1).to(
            dtype=x1_raw.dtype, device=x1_raw.device
        )
        occ_f = torch.zeros(batch_size, 1, h_x1, w_x1).to(
            dtype=x1_raw.dtype, device=x1_raw.device
        )
        occ_b = torch.zeros(batch_size, 1, h_x1, w_x1).to(
            dtype=x1_raw.dtype, device=x1_raw.device
        )

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):
            if l <= self.output_level:
                # warping
                if l == 0:
                    x2_warp = x2
                    x1_warp = x1
                else:
                    flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                    flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                    occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
                    occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
                    x2_warp = self.warping_layer(
                        x2, flow_f, height_im, width_im, self.div_flow
                    )
                    x1_warp = self.warping_layer(
                        x1, flow_b, height_im, width_im, self.div_flow
                    )

                # correlation
                out_corr_f = compute_cost_volume(x1, x2_warp, self.corr_params)
                out_corr_b = compute_cost_volume(x2, x1_warp, self.corr_params)

                out_corr_relu_f = self.leakyRELU(out_corr_f)
                out_corr_relu_b = self.leakyRELU(out_corr_b)

                if l != self.output_level:
                    x1_1by1 = self.conv_1x1[l](x1)
                    x2_1by1 = self.conv_1x1[l](x2)
                else:
                    x1_1by1 = x1
                    x2_1by1 = x2

                # concat and estimate flow
                flow_f = rescale_flow(
                    flow_f, self.div_flow, width_im, height_im, to_local=True
                )
                flow_b = rescale_flow(
                    flow_b, self.div_flow, width_im, height_im, to_local=True
                )

                x_intm_f, flow_res_f = self.flow_estimators(
                    torch.cat([out_corr_relu_f, x1_1by1, flow_f], dim=1)
                )
                x_intm_b, flow_res_b = self.flow_estimators(
                    torch.cat([out_corr_relu_b, x2_1by1, flow_b], dim=1)
                )
                flow_est_f = flow_f + flow_res_f
                flow_est_b = flow_b + flow_res_b

                flow_cont_f = flow_est_f + self.context_networks(
                    torch.cat([x_intm_f, flow_est_f], dim=1)
                )
                flow_cont_b = flow_est_b + self.context_networks(
                    torch.cat([x_intm_b, flow_est_b], dim=1)
                )

                # occ estimation
                x_intm_occ_f, occ_res_f = self.occ_estimators(
                    torch.cat([out_corr_relu_f, x1_1by1, occ_f], dim=1)
                )
                x_intm_occ_b, occ_res_b = self.occ_estimators(
                    torch.cat([out_corr_relu_b, x2_1by1, occ_b], dim=1)
                )
                occ_est_f = occ_f + occ_res_f
                occ_est_b = occ_b + occ_res_b

                occ_cont_f = occ_est_f + self.occ_context_networks(
                    torch.cat([x_intm_occ_f, occ_est_f], dim=1)
                )
                occ_cont_b = occ_est_b + self.occ_context_networks(
                    torch.cat([x_intm_occ_b, occ_est_b], dim=1)
                )

                # refinement
                img1_resize = upsample2d_as(x1_raw, flow_f, mode="bilinear")
                img2_resize = upsample2d_as(x2_raw, flow_b, mode="bilinear")
                flow_cont_f = rescale_flow(
                    flow_cont_f, self.div_flow, width_im, height_im, to_local=False
                )
                flow_cont_b = rescale_flow(
                    flow_cont_b, self.div_flow, width_im, height_im, to_local=False
                )
                img2_warp = self.warping_layer(
                    img2_resize, flow_cont_f, height_im, width_im, self.div_flow
                )
                img1_warp = self.warping_layer(
                    img1_resize, flow_cont_b, height_im, width_im, self.div_flow
                )

                # flow refine
                flow_f = self.refine_flow(
                    flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1
                )
                flow_b = self.refine_flow(
                    flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1
                )

                flow_f = rescale_flow(
                    flow_f, self.div_flow, width_im, height_im, to_local=False
                )
                flow_b = rescale_flow(
                    flow_b, self.div_flow, width_im, height_im, to_local=False
                )

                # occ refine
                x2_1by1_warp = self.warping_layer(
                    x2_1by1, flow_f, height_im, width_im, self.div_flow
                )
                x1_1by1_warp = self.warping_layer(
                    x1_1by1, flow_b, height_im, width_im, self.div_flow
                )

                occ_f = self.refine_occ(
                    occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp
                )
                occ_b = self.refine_occ(
                    occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp
                )

                flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

            else:
                flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                flows.append([flow_f, flow_b])

                x2_warp = self.warping_layer(
                    x2, flow_f, height_im, width_im, self.div_flow
                )
                x1_warp = self.warping_layer(
                    x1, flow_b, height_im, width_im, self.div_flow
                )
                flow_b_warp = self.warping_layer(
                    flow_b, flow_f, height_im, width_im, self.div_flow
                )
                flow_f_warp = self.warping_layer(
                    flow_f, flow_b, height_im, width_im, self.div_flow
                )

                if l != self.num_levels - 1:
                    x1_in = self.conv_1x1_1(x1)
                    x2_in = self.conv_1x1_1(x2)
                    x1_w_in = self.conv_1x1_1(x1_warp)
                    x2_w_in = self.conv_1x1_1(x2_warp)
                else:
                    x1_in = x1
                    x2_in = x2
                    x1_w_in = x1_warp
                    x2_w_in = x2_warp

                occ_f = self.occ_shuffle_upsample(
                    occ_f, torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp], dim=1)
                )
                occ_b = self.occ_shuffle_upsample(
                    occ_b, torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp], dim=1)
                )

                occs.append([occ_f, occ_b])

        flow_f_up = upsample2d_as(flow_f, x1_raw, mode="bilinear") * (
            1.0 / self.div_flow
        )
        flow_f_up = self.postprocess_predictions(flow_f_up, image_resizer, is_flow=True)
        flow_b_up = upsample2d_as(flow_b, x1_raw, mode="bilinear") * (
            1.0 / self.div_flow
        )
        flow_b_up = self.postprocess_predictions(flow_b_up, image_resizer, is_flow=True)
        occ_f_up = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")
        occ_f_up = self.postprocess_predictions(occ_f_up, image_resizer, is_flow=False)
        occ_b_up = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")
        occ_b_up = self.postprocess_predictions(occ_b_up, image_resizer, is_flow=False)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = flows
            outputs["occ_preds"] = occs
            outputs["flows"] = flow_f_up[:, None]
            outputs["occs"] = occ_f_up[:, None]
            outputs["flows_b"] = flow_b_up[:, None]
            outputs["occs_b"] = occ_b_up[:, None]
        else:
            outputs["occ_preds"] = occs
            outputs["flows"] = flow_f_up[:, None]
            outputs["occs"] = occ_f_up[:, None]
            outputs["flows_b"] = flow_b_up[:, None]
            outputs["occs_b"] = occ_b_up[:, None]
        return outputs


@register_model
@trainable
class irr_pwc(IRRPWC):
    pass



""" Twins
A PyTorch impl of : `Twins: Revisiting the Design of Spatial Attention in Vision Transformers`
    - https://arxiv.org/pdf/2104.13840.pdf

Code/weights from https://github.com/Meituan-AutoML/Twins, original copyright/license info below

"""

# --------------------------------------------------------
# Twins
# Copyright (c) 2021 Meituan
# Licensed under The Apache 2.0 License [see LICENSE for details]
# Written by Xinjie Li, Xiangxiang Chu
# --------------------------------------------------------
import math
from multiprocessing.sharedctypes import Value
import numpy as np
from copy import deepcopy
from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial

from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.layers import Mlp, DropPath, to_2tuple, trunc_normal_
from timm.models.fx_features import register_notrace_module
from timm.models.registry import register_model
from timm.models.vision_transformer import Attention
from timm.models.helpers import build_model_with_cfg


def get_2d_sincos_pos_embed(embed_dim, grid_size, cls_token=False, cossin=False):
    """
    grid_size: int of the grid height and width
    return:
    pos_embed: [grid_size*grid_size, embed_dim] or [1+grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    grid_h = np.arange(grid_size, dtype=np.float32)
    grid_w = np.arange(grid_size, dtype=np.float32)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size, grid_size])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid, cossin=cossin)
    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed


def get_2d_sincos_pos_embed_from_grid(embed_dim, grid, cossin=False):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(
        embed_dim // 2, grid[0], cossin=cossin
    )  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(
        embed_dim // 2, grid[1], cossin=cossin
    )  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1)  # (H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos, cossin=False):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    if cossin:
        emb = np.concatenate([emb_cos, emb_sin], axis=1)  # (M, D)
    else:
        emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb


# --------------------------------------------------------
# Interpolate position embeddings for high-resolution
# References:
# DeiT: https://github.com/facebookresearch/deit
# --------------------------------------------------------
def interpolate_pos_embed(model, checkpoint_model, remove_index=True):
    if "absolute_pos_embed" in checkpoint_model:
        pos_embed_checkpoint = checkpoint_model["absolute_pos_embed"]
        embedding_size = pos_embed_checkpoint.shape[-1]
        num_patches = model.patch_embed.num_patches
        num_extra_tokens = 0

        # height (== width) for the checkpoint position embedding
        orig_size = int((pos_embed_checkpoint.shape[-2] - num_extra_tokens) ** 0.5)
        # height (== width) for the new position embedding
        new_size = int(num_patches**0.5)
        # class_token and dist_token are kept unchanged
        if orig_size != new_size:
            print(
                "Position interpolate from %dx%d to %dx%d"
                % (orig_size, orig_size, new_size, new_size),
                flush=True,
            )
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            pos_tokens = pos_tokens.reshape(
                -1, orig_size, orig_size, embedding_size
            ).permute(0, 3, 1, 2)
            pos_tokens = torch.nn.functional.interpolate(
                pos_tokens,
                size=(new_size, new_size),
                mode="bicubic",
                align_corners=False,
            )
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model["absolute_pos_embed"] = new_pos_embed

    if "decoder_pos_embed" in checkpoint_model and hasattr(model, "decoder_pos_embed"):
        pos_embed_checkpoint = checkpoint_model["decoder_pos_embed"]
        embedding_size = pos_embed_checkpoint.shape[-1]
        num_patches = model.decoder_pos_embed.shape[-2]
        num_extra_tokens = 0

        # height (== width) for the checkpoint position embedding
        orig_size = int((pos_embed_checkpoint.shape[-2] - num_extra_tokens) ** 0.5)
        # height (== width) for the new position embedding
        new_size = int(num_patches**0.5)
        # class_token and dist_token are kept unchanged
        if orig_size != new_size:
            print(
                "Position interpolate from %dx%d to %dx%d"
                % (orig_size, orig_size, new_size, new_size),
                flush=True,
            )
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            pos_tokens = pos_tokens.reshape(
                -1, orig_size, orig_size, embedding_size
            ).permute(0, 3, 1, 2)
            pos_tokens = torch.nn.functional.interpolate(
                pos_tokens,
                size=(new_size, new_size),
                mode="bicubic",
                align_corners=False,
            )
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model["decoder_pos_embed"] = new_pos_embed

    # interpolate position bias table if needed
    relative_position_bias_table_keys = [
        k for k in checkpoint_model.keys() if "relative_position_bias_table" in k
    ]
    for table_key in relative_position_bias_table_keys:
        table_pretrained = checkpoint_model[table_key]
        if not table_key in model.state_dict():
            print(f"Key {table_key} not in model")
            continue
        table_current = model.state_dict()[table_key]
        L1, nH1 = table_pretrained.size()
        L2, nH2 = table_current.size()
        if nH1 != nH2:
            print(f"Error in loading {table_key}, pass")
        else:
            if L1 != L2:
                S1 = int(L1**0.5)
                S2 = int(L2**0.5)
                print(f"relative bias from {S1} to {S2}")
                table_pretrained_resized = F.interpolate(
                    table_pretrained.permute(1, 0).view(1, nH1, S1, S1),
                    size=(S2, S2),
                    mode="bicubic",
                )
                checkpoint_model[table_key] = table_pretrained_resized.view(
                    nH2, L2
                ).permute(1, 0)

    bias_index_keys = []
    for k in checkpoint_model.keys():
        if "relative_position_index" in k and remove_index:
            print(f"del {k} from pretrain ckpt")
            # do not load relative position index for workaround of changing input size
            bias_index_keys.append(k)
    for k in bias_index_keys:
        del checkpoint_model[k]

    bias_index_keys = []
    for k in checkpoint_model.keys():
        if "attn_mask" in k:
            print(f"del {k} from pretrain ckpt")
            bias_index_keys.append(k)
    for k in bias_index_keys:
        del checkpoint_model[k]


def _cfg(url="", **kwargs):
    return {
        "url": url,
        "num_classes": 1000,
        "input_size": (3, 224, 224),
        "pool_size": None,
        "crop_pct": 0.9,
        "interpolation": "bicubic",
        "fixed_input_size": True,
        "mean": IMAGENET_DEFAULT_MEAN,
        "std": IMAGENET_DEFAULT_STD,
        "first_conv": "patch_embeds.0.proj",
        "classifier": "head",
        **kwargs,
    }


default_cfgs = {
    "twins_pcpvt_small": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_small-e70e7e7a.pth",
    ),
    "twins_pcpvt_base": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_base-e5ecb09b.pth",
    ),
    "twins_pcpvt_large": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_large-d273f802.pth",
    ),
    "twins_svt_small": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_small-42e5f78c.pth",
    ),
    "twins_svt_base": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_base-c2265010.pth",
    ),
    "twins_svt_large": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_large-90f6aaa9.pth",
    ),
}

Size_ = Tuple[int, int]


@register_notrace_module  # reason: FX can't symbolically trace control flow in forward method
class LocallyGroupedAttn(nn.Module):
    """LSA: self attention within a group"""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, ws=1):
        assert ws != 1
        super(LocallyGroupedAttn, self).__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_, attn_mask=None):
        # There are two implementations for this function, zero padding or mask. We don't observe obvious difference for
        # both. You can choose any one, we recommend forward_padding because it's neat. However,
        # the masking implementation is more reasonable and accurate.
        B, N, C = x.shape
        H, W = size
        x = x.view(B, H, W, C)
        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        x = x.reshape(B, _h, self.ws, _w, self.ws, C).transpose(2, 3)
        qkv = (
            self.qkv(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 3, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]
        attn = (q @ k.transpose(-2, -1)) * self.scale

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class GlobalSubSampleAttn(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.kv = nn.Linear(dim, dim * 2, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, size: Size_, attn_mask=None):
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr is not None:
            x = x.permute(0, 2, 1).reshape(B, C, *size)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)
        kv = (
            self.kv(x)
            .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale

        _, _, L1, L2 = attn.shape

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio=1,
        ws=None,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.ws = ws
        if ws is None:
            self.attn = Attention(dim, num_heads, False, None, attn_drop, drop)
            raise ValueError("no")
        elif ws == 1:
            self.attn = GlobalSubSampleAttn(dim, num_heads, attn_drop, drop, sr_ratio)
        else:
            self.attn = LocallyGroupedAttn(dim, num_heads, attn_drop, drop, ws)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, size: Size_, attn_mask=None):
        x = x + self.drop_path(self.attn(self.norm1(x), size, attn_mask))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PosConv(nn.Module):
    # PEG  from https://arxiv.org/abs/2102.10882
    def __init__(self, in_chans, embed_dim=768, stride=1):
        super(PosConv, self).__init__()
        self.proj = nn.Sequential(
            nn.Conv2d(in_chans, embed_dim, 3, stride, 1, bias=True, groups=embed_dim),
        )
        self.stride = stride

    def forward(self, x, size: Size_):
        B, N, C = x.shape
        cnn_feat_token = x.transpose(1, 2).view(B, C, *size)
        x = self.proj(cnn_feat_token)
        if self.stride == 1:
            x += cnn_feat_token
        x = x.flatten(2).transpose(1, 2)
        return x

    def no_weight_decay(self):
        return ["proj.%d.weight" % i for i in range(4)]


class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        assert (
            img_size[0] % patch_size[0] == 0 and img_size[1] % patch_size[1] == 0
        ), f"img_size {img_size} should be divided by patch_size {patch_size}."
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x) -> Tuple[torch.Tensor, Size_]:
        B, C, H, W = x.shape

        x = self.proj(x).flatten(2).transpose(1, 2)
        x = self.norm(x)
        out_size = (H // self.patch_size[0], W // self.patch_size[1])

        return x, out_size


class Twins(nn.Module):
    """Twins Vision Transfomer (Revisiting Spatial Attention)

    Adapted from PVT (PyramidVisionTransformer) class at https://github.com/whai362/PVT.git
    """

    def __init__(
        self,
        img_size=224,
        patch_size=4,
        in_chans=3,
        num_classes=1000,
        global_pool="avg",
        embed_dims=(64, 128, 256, 512),
        num_heads=(1, 2, 4, 8),
        mlp_ratios=(4, 4, 4, 4),
        depths=(3, 4, 6, 3),
        sr_ratios=(8, 4, 2, 1),
        wss=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        block_cls=Block,
        norm_pix_loss=True,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.global_pool = global_pool
        self.depths = depths
        self.embed_dims = embed_dims
        self.num_features = embed_dims[-1]
        self.grad_checkpointing = False

        img_size = to_2tuple(img_size)
        prev_chs = in_chans
        self.patch_embeds = nn.ModuleList()
        self.pos_drops = nn.ModuleList()
        for i in range(len(depths)):
            self.patch_embeds.append(
                PatchEmbed(img_size, patch_size, prev_chs, embed_dims[i])
            )
            self.pos_drops.append(nn.Dropout(p=drop_rate))
            prev_chs = embed_dims[i]
            img_size = tuple(t // patch_size for t in img_size)
            patch_size = 2

        self.blocks = nn.ModuleList()
        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule
        cur = 0
        for k in range(len(depths)):
            _block = nn.ModuleList(
                [
                    block_cls(
                        dim=embed_dims[k],
                        num_heads=num_heads[k],
                        mlp_ratio=mlp_ratios[k],
                        drop=drop_rate,
                        attn_drop=attn_drop_rate,
                        drop_path=dpr[cur + i],
                        norm_layer=norm_layer,
                        sr_ratio=sr_ratios[k],
                        ws=1 if wss is None or i % 2 == 1 else wss[k],
                    )
                    for i in range(depths[k])
                ]
            )
            self.blocks.append(_block)
            cur += depths[k]

        # self.pos_block = nn.ModuleList([PosConv(embed_dim, embed_dim) for embed_dim in embed_dims])
        # self.pos_block = nn.ParameterList()
        for idx, embed_dim in enumerate(embed_dims):
            num_patches = [
                56**2,
                28**2,
                14**2,
                7**2,
            ]  # pretrain size is 224*224
            pe = nn.Parameter(
                torch.zeros(1, num_patches[idx], embed_dim), requires_grad=True
            )
            setattr(self, f"pos_block_{idx}", pe)
            _, N, C = pe.shape
            pos_embed = get_2d_sincos_pos_embed(C, int(N**0.5), cls_token=False)
            getattr(self, f"pos_block_{idx}").data.copy_(
                torch.from_numpy(pos_embed).float().unsqueeze(0)
            )
            # self.pos_block.append(pe)

        self.norm = norm_layer(self.num_features)

        # init weights
        self.apply(self._init_weights)
        # for pe in self.pos_block:
        #     _, N, C = pe.shape
        #     pos_embed = get_2d_sincos_pos_embed(C, int(N ** 0.5), cls_token=False)
        #     pe.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

    @torch.jit.ignore
    def no_weight_decay(self):
        return set(["pos_block." + n for n, p in self.pos_block.named_parameters()])

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def _scale_pe(self, pe, H, W):
        B, N, C = pe.shape
        h = int(N**0.5)
        assert h**2 == N
        pe = pe.view(B, h, h, C).permute(0, 3, 1, 2)
        pe = F.interpolate(pe, size=(H, W), mode="bicubic", align_corners=False)
        return pe.view(B, C, H, W).flatten(2).permute(0, 2, 1)

    def forward_features(self, x):
        B = x.shape[0]
        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(self.patch_embeds, self.pos_drops, self.blocks, self.pos_block)
        ):
            x, size = embed(x)
            x = drop(x)
            for j, blk in enumerate(blocks):
                x = blk(x, size)
                if j == 0:
                    # x = pos_blk(x, size)  # PEG here
                    x = x + self._scale_pe(pos_blk, *size)
            if i < len(self.depths) - 1:
                x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()
        x = self.norm(x)
        return x

    # def forward_head(self, x, pre_logits: bool = False):
    #     if self.global_pool == 'avg':
    #         x = x.mean(dim=1)
    #     return x if pre_logits else self.head(x)

    def forward(self, x):
        return self.forward_features(x)


def _create_twins(variant, pretrained=False, **kwargs):
    if kwargs.get("features_only", None):
        raise RuntimeError(
            "features_only not implemented for Vision Transformer models."
        )

    model = build_model_with_cfg(Twins, variant, pretrained, **kwargs)
    return model


@register_model
def _twins_svt_large_jihao(pretrained=False, **kwargs):
    model_kwargs = dict(
        patch_size=4,
        embed_dims=[128, 256, 512, 1024],
        num_heads=[4, 8, 16, 32],
        mlp_ratios=[4, 4, 4, 4],
        depths=[2, 2, 18, 2],
        wss=[7, 7, 7, 7],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    # return _create_twins('twins_svt_large', pretrained=pretrained, **model_kwargs)
    return Twins(**model_kwargs)



import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid[:, None] * i_loss).mean()

        return flow_loss


class CSFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/csflow-chairs-458a9436.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/csflow-things-ebdd403b.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/csflow-kitti-dc66357a.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        gen_fmap: bool = False,
        skip_encode: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.gen_fmap = gen_fmap
        self.skip_encode = skip_encode

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn="instance", dropout=dropout)

        self.cnet = BasicEncoder(
            output_dim=hdim + cdim, norm_fn="batch", dropout=dropout
        )

        self.strip_corr_block_v2 = StripCrossCorrMap_v2(in_chan=256, out_chan=256)
        self.update_block = BasicUpdateBlock(
            corr_levels=corr_levels, corr_radius=corr_radius, hidden_dim=hdim
        )

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids.
        flow = coords1 - coords0, Modified by Hao
        """
        N, C, H, W = img.shape

        coords0 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex
        combination."""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames."""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        if not self.skip_encode:
            # run the feature network
            fmap1, fmap2 = self.fnet([image1, image2])
        else:
            fmap1 = (image1 + 1) * 127.5
            fmap2 = (image2 + 1) * 127.5

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the context network
        if not self.skip_encode:
            cnet = self.cnet(image1)

            if not self.training:
                if self.gen_fmap:
                    return fmap1, fmap2, cnet
        else:
            cnet = inputs["images"][:, 2] * 255.0

        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        strip_coor_map, strip_corr_map_w, strip_corr_map_h = self.strip_corr_block_v2(
            [fmap1, fmap2]
        )
        corr_fn = CorrBlock_v2(fmap1, fmap2, strip_coor_map, radius=self.corr_radius)

        if not self.skip_encode:
            coords0, coords1 = self.initialize_flow(image1)
        else:
            b, c, h, w = fmap1.shape
            image1 = torch.zeros(b, c, 8 * h, 8 * w).cuda()
            coords0, coords1 = self.initialize_flow(image1)

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        flow_predictions = []

        # init flow with regression before GRU iters
        corr_w_act = torch.nn.functional.softmax(
            strip_corr_map_w, dim=3
        )  # B H1 W1 1 W2
        corr_h_act = torch.nn.functional.softmax(
            strip_corr_map_h, dim=4
        )  # B H1 W1 H2 1

        flo_v = corr_w_act.mul(strip_corr_map_w)  # B H1 W1 1 W2
        flo_u = corr_h_act.mul(strip_corr_map_h)  # B H1 W1 H2 1

        flow_v = torch.sum(flo_v, dim=4).squeeze(dim=3)  # B H1 W1
        flow_u = torch.sum(flo_u, dim=3).squeeze(dim=3)  # B H1 W1

        corr_init = torch.stack((flow_u, flow_v), dim=1)  # B 2 H1 W1

        coords1 = coords1.detach()
        coords1 = coords1 + corr_init

        # add loss
        flow_up = upflow8(coords1 - coords0)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
        flow_predictions.append(flow_up)

        for itr in range(self.iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume

            flow = coords1 - coords0
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


class StripCrossCorrMap_v2(nn.Module):
    """Strip Cross Corr Augmentation Module by Hao, version2.0"""

    def __init__(self, in_chan=256, out_chan=256, *args, **kwargs):
        super(StripCrossCorrMap_v2, self).__init__()
        self.conv1_1 = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)
        self.conv1_2 = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)
        self.conv2_1 = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)
        self.conv2_2 = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)

        self.init_weight()

    def forward(self, x):
        fmap1, fmap2 = x

        # vertical query map
        fmap1_w = self.conv1_1(fmap1)  # B, 64, H, W
        batchsize, c_middle, h, w = fmap1_w.size()
        fmap1_w = fmap1_w.view(batchsize, c_middle, -1)

        # horizontal query map
        fmap1_h = self.conv1_2(fmap1)  # B, 64, H, W
        batchsize, c_middle, h, w = fmap1_h.size()
        fmap1_h = fmap1_h.view(batchsize, c_middle, -1)

        # vertical striping map
        fmap2_w = self.conv2_1(fmap2)  # B, 64, H, W
        fmap2_w = F.avg_pool2d(fmap2_w, [h, 1])
        fmap2_w = fmap2_w.view(batchsize, c_middle, -1).permute(0, 2, 1)

        # horizontal striping map
        fmap2_h = self.conv2_2(fmap2)  # B, 64, H, W
        fmap2_h = F.avg_pool2d(fmap2_h, [1, w])
        fmap2_h = fmap2_h.view(batchsize, c_middle, -1).permute(0, 2, 1)

        # cross strip corr map
        strip_corr_map_w = (
            torch.bmm(fmap2_w, fmap1_w)
            .view(batchsize, w, h, w, 1)
            .permute(0, 2, 3, 4, 1)
        )  # B H1 W1 1 W2
        strip_corr_map_h = (
            torch.bmm(fmap2_h, fmap1_h)
            .view(batchsize, h, h, w, 1)
            .permute(0, 2, 3, 1, 4)
        )  # B H1 W1 H2 1

        return (
            (strip_corr_map_w + strip_corr_map_h).view(batchsize, h, w, 1, h, w),
            strip_corr_map_w,
            strip_corr_map_h,
        )

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None:
                    nn.init.constant_(ly.bias, 0)

    def get_params(self):
        wd_params, nowd_params = [], []
        for name, module in self.named_modules():
            if isinstance(module, (nn.Linear, nn.Conv2d)):
                wd_params.append(module.weight)
                if not module.bias is None:
                    nowd_params.append(module.bias)
            elif isinstance(module, torch.nn.BatchNorm2d):
                nowd_params += list(module.parameters())
        return wd_params, nowd_params


class ConvBNReLU(nn.Module):
    """Conv with BN and ReLU, used for Strip Corr Module"""

    def __init__(self, in_chan, out_chan, ks=3, stride=1, padding=1, *args, **kwargs):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(
            in_chan,
            out_chan,
            kernel_size=ks,
            stride=stride,
            padding=padding,
            bias=False,
        )
        self.bn = torch.nn.BatchNorm2d(out_chan)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

    def init_weight(self):
        for ly in self.children():
            if isinstance(ly, nn.Conv2d):
                nn.init.kaiming_normal_(ly.weight, a=1)
                if not ly.bias is None:
                    nn.init.constant_(ly.bias, 0)


class BasicUpdateBlock(nn.Module):
    """Modified by Hao, support for CSFlow"""

    def __init__(self, corr_levels: int, corr_radius: int, hidden_dim: int = 128):
        super(BasicUpdateBlock, self).__init__()
        self.encoder = BasicMotionEncoder_v2(
            corr_levels=corr_levels, corr_radius=corr_radius
        )
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


def pool2x(x):
    return F.avg_pool2d(x, 3, stride=2, padding=1)


def interp(x, dest):
    interp_args = {"mode": "bilinear", "align_corners": True}
    return F.interpolate(x, dest.shape[2:], **interp_args)


class BasicEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        from torch.nn.modules.utils import _pair

        super(BasicEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64, stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)

        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class CorrBlock_v2:
    """Corr Block, modified by Hao, concat SC with 4D corr"""

    def __init__(self, fmap1, fmap2, strip_coor_map=None, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid = []

        # all pairs correlation
        corr = CorrBlock_v2.corr(fmap1, fmap2)

        if strip_coor_map is not None:
            # strip correlation augmentation with concat
            corr = torch.cat((corr, strip_coor_map), dim=3)

        batch, h1, w1, dim, h2, w2 = corr.shape
        corr = corr.reshape(batch * h1 * w1, dim, h2, w2)

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            dx = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            dy = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2).contiguous()

    @staticmethod
    def corr(fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(fmap1.transpose(1, 2), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr / torch.sqrt(torch.tensor(dim))


def bilinear_sampler(img, coords, mode="bilinear", mask=False):
    """Wrapper for grid_sample, uses pixel coordinates."""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.to(dtype=coords.dtype)

    return img


def coords_grid(batch, ht, wd, dtype, device):
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
        indexing="ij",
    )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].repeat(batch, 1, 1, 1)


def upflow8(flow, mode="bilinear"):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return 8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(ResidualBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BottleneckBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(BottleneckBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_planes, planes // 4, kernel_size=1, padding=0)
        self.conv2 = nn.Conv2d(
            planes // 4, planes // 4, kernel_size=3, padding=1, stride=stride
        )
        self.conv3 = nn.Conv2d(planes // 4, planes, kernel_size=1, padding=0)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm4 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes // 4)
            self.norm2 = nn.BatchNorm2d(planes // 4)
            self.norm3 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes // 4)
            self.norm2 = nn.InstanceNorm2d(planes // 4)
            self.norm3 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()
            if not stride == 1:
                self.norm4 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm4
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))
        y = self.relu(self.norm3(self.conv3(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BasicMotionEncoder_v2(nn.Module):
    """Get Motion Feature from CSFlow, by Hao"""

    def __init__(self, corr_levels: int, corr_radius: int):
        super(BasicMotionEncoder_v2, self).__init__()
        # double cor_plances due to concat aug
        cor_planes = 2 * (corr_levels * (2 * corr_radius + 1) ** 2)
        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 128 - 2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class SepConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convr1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convq1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )

        self.convz2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convr2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convq2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )

    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        return h


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class ConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(ConvGRU, self).__init__()
        self.convz = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convr = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convq = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = torch.sigmoid(self.convz(hx))
        r = torch.sigmoid(self.convr(hx))
        q = torch.tanh(self.convq(torch.cat([r * h, x], dim=1)))

        h = (1 - z) * h + z * q
        return h


@register_model
@trainable
class csflow(CSFlow):
    pass



#!/usr/bin/env python

import os

from setuptools import setup, find_packages
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

this_file = os.path.dirname(__file__)

setup(
    name="quadtree_attention_package",
    ext_modules=[
        CUDAExtension(
            "score_computation_cuda",
            [
                "QuadtreeAttention/src/score_computation.cpp",
                "QuadtreeAttention/src/score_computation_kernal.cu",
            ],
            extra_compile_args={"cxx": ["-g"], "nvcc": ["-O2"]},
        ),
        CUDAExtension(
            "value_aggregation_cuda",
            [
                "QuadtreeAttention/src/value_aggregation.cpp",
                "QuadtreeAttention/src/value_aggregation_kernel.cu",
            ],
            extra_compile_args={"cxx": ["-g"], "nvcc": ["-O2"]},
        ),
    ],
    cmdclass={"build_ext": BuildExtension},
)



from .conv2d_same import Conv2dSame, conv2d_same
from .drop import DropBlock2d, DropPath, drop_block_2d, drop_path
from .helpers import to_ntuple, to_2tuple, to_3tuple, to_4tuple, make_divisible
from .linear import Linear
from .mlp import Mlp, GluMlp, GatedMlp
from .patch_embed import PatchEmbed
from .weight_init import trunc_normal_, variance_scaling_, lecun_normal_

# from .activations import *
# from .adaptive_avgmax_pool import \
#     adaptive_avgmax_pool2d, select_adaptive_pool2d, AdaptiveAvgMaxPool2d, SelectAdaptivePool2d
# from .blur_pool import BlurPool2d
# from .classifier import ClassifierHead, create_classifier
# from .cond_conv2d import CondConv2d, get_condconv_initializer
# from .config import is_exportable, is_scriptable, is_no_jit, set_exportable, set_scriptable, set_no_jit,\
#     set_layer_config
# from .conv_bn_act import ConvBnAct
# from .create_act import create_act_layer, get_act_layer, get_act_fn
# from .create_attn import get_attn, create_attn
# from .create_conv2d import create_conv2d
# from .create_norm_act import get_norm_act_layer, create_norm_act, convert_norm_act
# from .eca import EcaModule, CecaModule, EfficientChannelAttn, CircularEfficientChannelAttn
# from .evo_norm import EvoNormBatch2d, EvoNormSample2d
# from .gather_excite import GatherExcite
# from .global_context import GlobalContext
# from .inplace_abn import InplaceAbn
# from .involution import Involution
# from .mixed_conv2d import MixedConv2d
# from .non_local_attn import NonLocalAttn, BatNonLocalAttn
# from .norm import GroupNorm, LayerNorm2d
# from .norm_act import BatchNormAct2d, GroupNormAct
# from .padding import get_padding, get_same_padding, pad_same
# from .pool2d_same import AvgPool2dSame, create_pool2d
# from .squeeze_excite import SEModule, SqueezeExcite, EffectiveSEModule, EffectiveSqueezeExcite
# from .selective_kernel import SelectiveKernel
# from .separable_conv import SeparableConv2d, SeparableConvBnAct
# from .space_to_depth import SpaceToDepthModule
# from .split_attn import SplitAttn
# from .split_batchnorm import SplitBatchNorm2d, convert_splitbn_model
# from .std_conv import StdConv2d, StdConv2dSame, ScaledStdConv2d, ScaledStdConv2dSame
# from .test_time_pool import TestTimePoolHead, apply_test_time_pool



import torch
import torch.nn.functional as F
from .utils import coords_grid

try:
    import alt_cuda_corr
except:
    # alt_cuda_corr is not compiled
    pass


class PathMatch:
    def __init__(self, fmap1, fmap2):
        self.map1 = fmap1
        self.map2 = fmap2
        self.N, self.C, self.H, self.W = fmap1.shape
        self.single_planes = self.C // 2
        self.coords = coords_grid(
            self.N, self.H, self.W, dtype=fmap1.dtype, device=fmap1.device
        )
        self.shift_map2 = self.get_inv_patch_map(fmap2)
        self.view_map1 = fmap1.view(self.N, self.C // 2, 2, 1, self.H, self.W)

    def get_inv_patch_map(self, fmap2):
        #
        r = 1
        fmap2_tl = F.pad(fmap2, [r, 0, r, 0], mode="replicate")[
            :, :, 0 : self.H, 0 : self.W
        ]
        fmap2_tr = F.pad(fmap2, [0, r, r, 0], mode="replicate")[:, :, 0 : self.H, r:]
        fmap2_dl = F.pad(fmap2, [r, 0, 0, r], mode="replicate")[:, :, r:, 0 : self.W]
        fmap2_dr = F.pad(fmap2, [0, r, 0, r], mode="replicate")[:, :, r:, r:]
        return torch.cat((fmap2, fmap2_tl, fmap2_tr, fmap2_dl, fmap2_dr), dim=1)

    def warp(self, coords, image, h, w):
        coords[:, 0, :, :] = 2.0 * coords[:, 0, :, :].clone() / max(self.W - 1, 1) - 1.0
        coords[:, 1, :, :] = 2.0 * coords[:, 1, :, :].clone() / max(self.H - 1, 1) - 1.0

        coords = coords.permute(0, 2, 3, 1)
        output = F.grid_sample(image, coords, align_corners=True, padding_mode="border")
        return output

    def search(self, flow, scale=1):
        corrs = []
        temp_coord = self.coords + flow
        map2_warp = self.warp(temp_coord, self.map2, self.H, self.W)
        padd_map2 = F.pad(map2_warp, [2, 2, 2, 2], mode="replicate")
        # #current
        # random search
        for i in range(5):
            for j in range(5):
                map2 = padd_map2[:, :, j : j + self.H, i : i + self.W]
                cost = torch.mean(self.map1 * map2, dim=1, keepdim=True)
                corrs.append(cost)

        out_corrs = torch.cat(corrs, dim=1)
        return out_corrs

    def inverse_propagation(self, flow):
        corrs = []
        temp_coord = self.coords + flow
        map2_warp = self.warp(temp_coord, self.shift_map2, self.H, self.W)

        map2_warp = map2_warp.view(self.N, self.C // 2, 2, 5, self.H, self.W)
        corr = torch.mean(map2_warp * self.view_map1, dim=1)
        corr = corr.view(self.N, 10, self.H, self.W)
        return corr

    def __call__(self, flow, is_search=True):
        if is_search:
            out_corrs = self.search(flow)
        else:
            out_corrs = self.inverse_propagation(flow)
        return out_corrs



import torch
from torch import nn, einsum
from einops import rearrange


class RelPosEmb(nn.Module):
    def __init__(self, max_pos_size, dim_head):
        super().__init__()
        self.rel_height = nn.Embedding(2 * max_pos_size - 1, dim_head)
        self.rel_width = nn.Embedding(2 * max_pos_size - 1, dim_head)

        deltas = torch.arange(max_pos_size).view(1, -1) - torch.arange(
            max_pos_size
        ).view(-1, 1)
        rel_ind = deltas + max_pos_size - 1
        self.register_buffer("rel_ind", rel_ind)

    def forward(self, q):
        batch, heads, h, w, c = q.shape
        height_emb = self.rel_height(self.rel_ind[:h, :h].reshape(-1))
        width_emb = self.rel_width(self.rel_ind[:w, :w].reshape(-1))

        height_emb = rearrange(height_emb, "(x u) d -> x u () d", x=h)
        width_emb = rearrange(width_emb, "(y v) d -> y () v d", y=w)

        height_score = einsum("b h x y d, x u v d -> b h x y u v", q, height_emb)
        width_score = einsum("b h x y d, y u v d -> b h x y u v", q, width_emb)

        return height_score + width_score


class Attention(nn.Module):
    def __init__(
        self,
        *,
        dim,
        max_pos_size=100,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_qk = nn.Conv2d(dim, inner_dim * 2, 1, bias=False)

        self.pos_emb = RelPosEmb(max_pos_size, dim_head)

    def forward(self, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        q, k = self.to_qk(fmap).chunk(2, dim=1)

        q, k = map(lambda t: rearrange(t, "b (h d) x y -> b h x y d", h=heads), (q, k))
        q = self.scale * q

        sim = einsum("b h x y d, b h u v d -> b h x y u v", q, k)

        sim = rearrange(sim, "b h x y u v -> b h (x y) (u v)")
        attn = sim.softmax(dim=-1)

        return attn


class QKEncoder(nn.Module):
    def __init__(self, in_dim, outdim):
        super().__init__()
        self.query_proj = nn.Conv2d(in_dim, outdim, kernel_size=3, padding=1)
        self.key_proj = nn.Conv2d(in_dim, outdim, kernel_size=3, padding=1)
        # shrinkage
        self.d_proj = nn.Conv2d(in_dim, 1, kernel_size=3, padding=1)
        # selection
        self.e_proj = nn.Conv2d(in_dim, outdim, kernel_size=3, padding=1)

        nn.init.orthogonal_(self.query_proj.weight.data)
        nn.init.zeros_(self.query_proj.bias.data)
        nn.init.orthogonal_(self.key_proj.weight.data)
        nn.init.zeros_(self.key_proj.bias.data)

    def forward(self, x, need_s, need_e):
        shrinkage = self.d_proj(x) ** 2 + 1 if (need_s) else None
        selection = torch.sigmoid(self.e_proj(x)) if (need_e) else None

        return self.query_proj(x), self.key_proj(x), shrinkage, selection


class Aggregate(nn.Module):
    def __init__(
        self,
        dim,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_v = nn.Conv2d(dim, inner_dim, 1, bias=False)

        self.gamma = nn.Parameter(torch.zeros(1))

        if dim != inner_dim:
            self.project = nn.Conv2d(inner_dim, dim, 1, bias=False)
        else:
            self.project = None

    def forward(self, attn, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        v = self.to_v(fmap)
        v = rearrange(v, "b (h d) x y -> b h (x y) d", h=heads)
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)

        if self.project is not None:
            out = self.project(out)

        out = fmap + self.gamma * out

        return out



from loguru import logger
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable, ptlflow_trained
from ptlflow.utils.utils import forward_interpolate_batch
from .update import BasicUpdateBlock, SmallUpdateBlock
from .extractor import BasicEncoder, SmallEncoder
from .corr import get_corr_block
from .utils import coords_grid, upflow8
from ..base_model.base_model import BaseModel

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class RAFT(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-chairs-590f38f7.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-things-802bbcfd.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-sintel-fb44381e.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-kitti-3a831a4b.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        alternate_corr: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.alternate_corr = alternate_corr

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn="instance", dropout=dropout)
        self.cnet = BasicEncoder(
            output_dim=hdim + cdim, norm_fn="batch", dropout=dropout
        )
        self.update_block = BasicUpdateBlock(
            self.corr_levels, self.corr_radius, hidden_dim=hdim
        )

        self.has_trained_on_ptlflow = True

        if self.alternate_corr and alt_cuda_corr is None:
            logger.warning(
                "!!! alt_cuda_corr is not compiled! The slower IterativeCorrBlock will be used instead !!!"
            )

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])

        corr_fn = get_corr_block(
            fmap1=fmap1,
            fmap2=fmap2,
            radius=self.corr_radius,
            num_levels=self.corr_levels,
            alternate_corr=self.alternate_corr,
        )

        # run the context network
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        flow_predictions = []
        for itr in range(self.iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume

            flow = coords1 - coords0
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


class RAFTSmall(RAFT):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft_small-things-b7d9f997.ckpt"
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 3,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        alternate_corr: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels=corr_levels,
            corr_radius=corr_radius,
            dropout=dropout,
            gamma=gamma,
            max_flow=max_flow,
            iters=iters,
            alternate_corr=alternate_corr,
            **kwargs,
        )
        self.hidden_dim = hdim = 96
        self.context_dim = cdim = 64

        # feature network, context network, and update block
        self.fnet = SmallEncoder(
            output_dim=128, norm_fn="instance", dropout=self.dropout
        )
        self.cnet = SmallEncoder(
            output_dim=hdim + cdim, norm_fn="none", dropout=self.dropout
        )
        self.update_block = SmallUpdateBlock(corr_levels, corr_radius, hidden_dim=hdim)


@register_model
@trainable
@ptlflow_trained
class raft(RAFT):
    pass


@register_model
@trainable
@ptlflow_trained
class raft_small(RAFTSmall):
    pass



import torch
from torch import nn, einsum
from einops import rearrange


class Attention(nn.Module):
    def __init__(
        self,
        *,
        dim,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_qk = nn.Conv2d(dim, inner_dim * 2, 1, bias=False)

    def forward(self, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        q, k = self.to_qk(fmap).chunk(2, dim=1)

        q, k = map(lambda t: rearrange(t, "b (h d) x y -> b h x y d", h=heads), (q, k))
        q = self.scale * q

        sim = einsum("b h x y d, b h u v d -> b h x y u v", q, k)

        sim = rearrange(sim, "b h x y u v -> b h (x y) (u v)")
        attn = sim.softmax(dim=-1)

        return attn


class Aggregate(nn.Module):
    def __init__(
        self,
        dim,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_v = nn.Conv2d(dim, inner_dim, 1, bias=False)

        self.gamma = nn.Parameter(torch.zeros(1))

        if dim != inner_dim:
            self.project = nn.Conv2d(inner_dim, dim, 1, bias=False)
        else:
            self.project = None

    def forward(self, attn, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        v = self.to_v(fmap)
        v = rearrange(v, "b (h d) x y -> b h (x y) d", h=heads)
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)

        if self.project is not None:
            out = self.project(out)

        out = fmap + self.gamma * out

        return out



# freda (todo) :

import torch.nn as nn
import torch
import numpy as np

try:
    from spatial_correlation_sampler import spatial_correlation_sample
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        iter_spatial_correlation_sample as spatial_correlation_sample,
    )


def conv(batchNorm, in_planes, out_planes, kernel_size=3, stride=1):
    if batchNorm:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=(kernel_size - 1) // 2,
                bias=False,
            ),
            nn.BatchNorm2d(out_planes),
            nn.LeakyReLU(0.1, inplace=True),
        )
    else:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=(kernel_size - 1) // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.1, inplace=True),
        )


def i_conv(batchNorm, in_planes, out_planes, kernel_size=3, stride=1, bias=True):
    if batchNorm:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=(kernel_size - 1) // 2,
                bias=bias,
            ),
            nn.BatchNorm2d(out_planes),
        )
    else:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=(kernel_size - 1) // 2,
                bias=bias,
            ),
        )


def predict_flow(in_planes):
    return nn.Conv2d(in_planes, 2, kernel_size=3, stride=1, padding=1, bias=True)


def deconv(in_planes, out_planes):
    return nn.Sequential(
        nn.ConvTranspose2d(
            in_planes, out_planes, kernel_size=4, stride=2, padding=1, bias=True
        ),
        nn.LeakyReLU(0.1, inplace=True),
    )


class tofp16(nn.Module):
    def __init__(self):
        super(tofp16, self).__init__()

    def forward(self, input):
        return input.half()


class tofp32(nn.Module):
    def __init__(self):
        super(tofp32, self).__init__()

    def forward(self, input):
        return input.float()


def init_deconv_bilinear(weight):
    f_shape = weight.size()
    heigh, width = f_shape[-2], f_shape[-1]
    f = np.ceil(width / 2.0)
    c = (2 * f - 1 - f % 2) / (2.0 * f)
    bilinear = np.zeros([heigh, width])
    for x in range(width):
        for y in range(heigh):
            value = (1 - abs(x / f - c)) * (1 - abs(y / f - c))
            bilinear[x, y] = value
    weight.data.fill_(0.0)
    for i in range(f_shape[0]):
        for j in range(f_shape[1]):
            weight.data[i, j, :, :] = torch.from_numpy(bilinear)


def save_grad(grads, name):
    def hook(grad):
        grads[name] = grad

    return hook


"""
def save_grad(grads, name):
    def hook(grad):
        grads[name] = grad
    return hook
import torch
from channelnorm_package.modules.channelnorm import ChannelNorm 
model = ChannelNorm().cuda()
grads = {}
a = 100*torch.autograd.Variable(torch.randn((1,3,5,5)).cuda(), requires_grad=True)
a.register_hook(save_grad(grads, 'a'))
b = model(a)
y = torch.mean(b)
y.backward()
"""


def correlate(input1, input2):
    out_corr = spatial_correlation_sample(
        input1,
        input2,
        kernel_size=1,
        patch_size=21,
        stride=1,
        padding=0,
        dilation_patch=2,
    )
    # collate dimensions 1 and 2 in order to be treated as a
    # regular 4D tensor
    b, ph, pw, h, w = out_corr.size()
    out_corr = out_corr.view(b, ph * pw, h, w) / input1.size(1)
    return out_corr



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import torch
import torch.nn as nn

from .conv import Conv2dBlock
from .cgu import CGUStage
from .norm import LayerNorm2d


class FlowHead(nn.Module):
    def __init__(
        self, input_dim=128, hidden_dim=256, activation_function=None, info_pred=False
    ):
        super(FlowHead, self).__init__()
        self.conv1 = Conv2dBlock(input_dim, hidden_dim, 3, padding=1)
        out_ch = 6 if info_pred else 2
        self.conv2 = Conv2dBlock(hidden_dim, out_ch, 3, padding=1)

        act = nn.ReLU if activation_function is None else activation_function
        self.act = act(inplace=True)

    def forward(self, x):
        return self.conv2(self.act(self.conv1(x)))


class ConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(ConvGRU, self).__init__()
        self.convz = Conv2dBlock(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convr = Conv2dBlock(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convq = Conv2dBlock(hidden_dim + input_dim, hidden_dim, 3, padding=1)

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = self.convz(hx)
        z = torch.sigmoid(z)

        r = self.convr(hx)
        r = torch.sigmoid(r)

        q = self.convq(torch.cat([r * h, x], dim=1))
        q = torch.tanh(q)

        h = (1 - z) * h + z * q
        return h


class CGUGRU(nn.Module):
    def __init__(
        self,
        hidden_dim=128,
        input_dim=192 + 128,
        activation_function=None,
        norm_layer=LayerNorm2d,
        depth=2,
        mlp_ratio=4,
        mlp_use_dw_conv=True,
        mlp_dw_kernel_size=7,
        mlp_in_kernel_size=1,
        mlp_out_kernel_size=1,
        layer_scale_init_value=1e-2,
    ):
        super(CGUGRU, self).__init__()

        self.convz = CGUStage(
            hidden_dim + input_dim,
            hidden_dim,
            stride=1,
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            use_cross=False,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            layer_scale_init_value=layer_scale_init_value,
        )
        self.convr = CGUStage(
            hidden_dim + input_dim,
            hidden_dim,
            stride=1,
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            use_cross=False,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            layer_scale_init_value=layer_scale_init_value,
        )
        self.convq = CGUStage(
            hidden_dim + input_dim,
            hidden_dim,
            stride=1,
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            use_cross=False,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            layer_scale_init_value=layer_scale_init_value,
        )

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = self.convz(hx)
        z = torch.sigmoid(z)

        r = self.convr(hx)
        r = torch.sigmoid(r)

        q = self.convq(torch.cat([r * h, x], dim=1))
        q = torch.tanh(q)

        h = (1 - z) * h + z * q
        return h


class ConvexMask(nn.Module):
    def __init__(self, net_chs, pred_stride, activation_function=None):
        super(ConvexMask, self).__init__()
        self.conv1 = Conv2dBlock(net_chs, net_chs * 2, 3, padding=1)
        self.conv2 = Conv2dBlock(net_chs * 2, pred_stride**2 * 9, 1, padding=0)

        act = nn.ReLU if activation_function is None else activation_function
        self.act = act(inplace=True)

    def forward(self, x):
        x = self.conv1(x)
        x = self.act(x)
        x = self.conv2(x)
        return x


class MotionEncoder(nn.Module):
    def __init__(
        self,
        dec_motenc_corr_hidden_chs: int,
        dec_motenc_corr_out_chs: int,
        dec_motenc_flow_hidden_chs: int,
        dec_motenc_flow_out_chs: int,
        corr_levels: int,
        corr_range: int,
        dec_flow_kernel_size: int,
        dec_motion_chs: int,
        activation_function: callable,
    ):
        super(MotionEncoder, self).__init__()

        c_hidden = dec_motenc_corr_hidden_chs
        c_out = dec_motenc_corr_out_chs
        f_hidden = dec_motenc_flow_hidden_chs
        f_out = dec_motenc_flow_out_chs

        cor_planes = corr_levels * (2 * corr_range + 1) ** 2
        self.convc1 = Conv2dBlock(cor_planes, c_hidden, 1, padding=0)
        self.convc2 = Conv2dBlock(c_hidden, c_out, 3, padding=1)

        self.convf1 = Conv2dBlock(
            2,
            f_hidden,
            dec_flow_kernel_size,
            padding=dec_flow_kernel_size // 2,
        )
        self.convf2 = Conv2dBlock(f_hidden, f_out, 3, padding=1)

        in_ch = f_out + c_out
        out_ch = dec_motion_chs - 2
        self.conv = Conv2dBlock(in_ch, out_ch, 3, padding=1)

        act = nn.ReLU if activation_function is None else activation_function
        self.act = act(inplace=True)

    def forward(self, flow, corr):
        cor = self.act(self.convc1(corr))
        cor = self.act(self.convc2(cor))

        outs = [cor]

        flo = self.act(self.convf1(flow))
        flo = self.act(self.convf2(flo))
        outs.append(flo)

        outs = torch.cat(outs, dim=1)
        out = self.act(self.conv(outs))
        out_t = [out, flow]
        return torch.cat(out_t, dim=1)


class UpdateBlock(nn.Module):
    def __init__(
        self,
        dec_motenc_corr_hidden_chs: int,
        dec_motenc_corr_out_chs: int,
        dec_motenc_flow_hidden_chs: int,
        dec_motenc_flow_out_chs: int,
        corr_levels: int,
        corr_range: int,
        dec_flow_kernel_size: int,
        dec_motion_chs: int,
        activation_function: callable,
        net_chs_fixed: int,
        inp_chs_fixed: int,
        dec_gru_norm_layer: callable,
        dec_gru_depth: int,
        dec_gru_iters: int,
        dec_gru_mlp_ratio: float,
        cgu_mlp_use_dw_conv: bool,
        cgu_mlp_dw_kernel_size: int,
        dec_gru_mlp_in_kernel_size: int,
        dec_gru_mlp_out_kernel_size: int,
        cgu_layer_scale_init_value: float,
        dec_flow_head_chs: int,
        loss: str,
        use_upsample_mask: bool,
        upmask_gradient_scale: float,
    ):
        super(UpdateBlock, self).__init__()
        self.use_upsample_mask = use_upsample_mask
        self.upmask_gradient_scale = upmask_gradient_scale

        self.encoder = MotionEncoder(
            dec_motenc_corr_hidden_chs=dec_motenc_corr_hidden_chs,
            dec_motenc_corr_out_chs=dec_motenc_corr_out_chs,
            dec_motenc_flow_hidden_chs=dec_motenc_flow_hidden_chs,
            dec_motenc_flow_out_chs=dec_motenc_flow_out_chs,
            corr_levels=corr_levels,
            corr_range=corr_range,
            dec_flow_kernel_size=dec_flow_kernel_size,
            dec_motion_chs=dec_motion_chs,
            activation_function=activation_function,
        )

        self.gru_list = nn.ModuleList(
            [
                CGUGRU(
                    hidden_dim=net_chs_fixed,
                    input_dim=dec_motion_chs + inp_chs_fixed,
                    activation_function=activation_function,
                    norm_layer=dec_gru_norm_layer,
                    depth=dec_gru_depth,
                    mlp_ratio=dec_gru_mlp_ratio,
                    mlp_use_dw_conv=cgu_mlp_use_dw_conv,
                    mlp_dw_kernel_size=cgu_mlp_dw_kernel_size,
                    mlp_in_kernel_size=dec_gru_mlp_in_kernel_size,
                    mlp_out_kernel_size=dec_gru_mlp_out_kernel_size,
                    layer_scale_init_value=cgu_layer_scale_init_value,
                )
                for _ in range(dec_gru_iters)
            ]
        )

        self.flow_head = FlowHead(
            net_chs_fixed,
            hidden_dim=dec_flow_head_chs,
            activation_function=activation_function,
            info_pred=(loss == "laplace"),
        )

        if use_upsample_mask:
            pred_stride = 8
            self.mask = ConvexMask(
                net_chs_fixed,
                pred_stride,
                activation_function=activation_function,
            )

    def forward(self, net, inp, corr, flow):
        motion_features = self.encoder(flow, corr)

        inp = torch.cat([inp, motion_features], dim=1)

        for gru in self.gru_list:
            net = gru(net, inp)

        delta_flow = self.flow_head(net)

        mask = None
        if self.use_upsample_mask:
            mask = self.upmask_gradient_scale * self.mask(net)

        return delta_flow, net, mask



from importlib.metadata import version

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from . import backbone_v7
from . import transformer
from . import matching
from . import corr
from . import refine
from . import upsample
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        weights = [0.2, 1]
        for i in range(n_predictions):
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += weights[i] * (valid * i_loss).mean()

        return flow_loss


class NeuFlow2(BaseModel):
    pretrained_checkpoints = {
        "mixed": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/neuflow2-mixed-acac1a70.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/neuflow2-sintel-15c625f8.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/neuflow2-things-6ed47437.ckpt",
    }

    def __init__(
        self,
        gamma: float = 0.8,
        max_flow: float = 400,
        feature_dim_s16: int = 128,
        context_dim_s16: int = 64,
        iter_context_dim_s16: int = 64,
        feature_dim_s8: int = 128,
        context_dim_s8: int = 64,
        iter_context_dim_s8: int = 64,
        feature_dim_s1: int = 128,
        iters_s16: int = 1,
        iters_s8: int = 8,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=16, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.gamma = gamma
        self.max_flow = max_flow
        self.feature_dim_s16 = feature_dim_s16
        self.context_dim_s16 = context_dim_s16
        self.iter_context_dim_s16 = iter_context_dim_s16
        self.feature_dim_s8 = feature_dim_s8
        self.context_dim_s8 = context_dim_s8
        self.iter_context_dim_s8 = iter_context_dim_s8
        self.feature_dim_s1 = feature_dim_s1
        self.iters_s16 = iters_s16
        self.iters_s8 = iters_s8

        if not version("torch").startswith("2"):
            raise ImportError(
                f"NeuFlow 2 requires torch 2.X. Your current version is {version('torch')}"
            )

        self.backbone = backbone_v7.CNNEncoder(
            self.feature_dim_s16,
            self.context_dim_s16,
            self.feature_dim_s8,
            self.context_dim_s8,
        )

        self.cross_attn_s16 = transformer.FeatureAttention(
            self.feature_dim_s16 + self.context_dim_s16,
            num_layers=2,
            ffn=True,
            ffn_dim_expansion=1,
            post_norm=True,
        )

        self.matching_s16 = matching.Matching()

        # self.flow_attn_s16 = transformer.FlowAttention(self.feature_dim_s16)

        self.corr_block_s16 = corr.CorrBlock(radius=4, levels=1)
        self.corr_block_s8 = corr.CorrBlock(radius=4, levels=1)

        self.merge_s8 = torch.nn.Sequential(
            torch.nn.Conv2d(
                self.feature_dim_s16 + self.feature_dim_s8,
                self.feature_dim_s8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            torch.nn.GELU(),
            torch.nn.Conv2d(
                self.feature_dim_s8,
                self.feature_dim_s8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            torch.nn.BatchNorm2d(self.feature_dim_s8),
        )

        self.context_merge_s8 = torch.nn.Sequential(
            torch.nn.Conv2d(
                self.context_dim_s16 + self.context_dim_s8,
                self.context_dim_s8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            torch.nn.GELU(),
            torch.nn.Conv2d(
                self.context_dim_s8,
                self.context_dim_s8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            torch.nn.BatchNorm2d(self.context_dim_s8),
        )

        self.refine_s16 = refine.Refine(
            self.context_dim_s16,
            self.iter_context_dim_s16,
            num_layers=5,
            levels=1,
            radius=4,
            inter_dim=128,
        )
        self.refine_s8 = refine.Refine(
            self.context_dim_s8,
            self.iter_context_dim_s8,
            num_layers=5,
            levels=1,
            radius=4,
            inter_dim=96,
        )

        self.conv_s8 = backbone_v7.ConvBlock(
            3, self.feature_dim_s1, kernel_size=8, stride=8, padding=0
        )
        self.upsample_s8 = upsample.UpSample(self.feature_dim_s1, upsample_factor=8)

        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)

        self.has_init_bhwd = False

    def init_bhwd(self, batch_size, height, width, device, amp=True):
        self.backbone.init_bhwd(batch_size * 2, height // 16, width // 16, device, amp)

        self.matching_s16.init_bhwd(batch_size, height // 16, width // 16, device, amp)

        self.corr_block_s16.init_bhwd(
            batch_size, height // 16, width // 16, device, amp
        )
        self.corr_block_s8.init_bhwd(batch_size, height // 8, width // 8, device, amp)

        self.refine_s16.init_bhwd(batch_size, height // 16, width // 16, device, amp)
        self.refine_s8.init_bhwd(batch_size, height // 8, width // 8, device, amp)

        self.init_iter_context_s16 = torch.zeros(
            batch_size,
            self.iter_context_dim_s16,
            height // 16,
            width // 16,
            device=device,
            dtype=torch.half if amp else torch.float,
        )
        self.init_iter_context_s8 = torch.zeros(
            batch_size,
            self.iter_context_dim_s8,
            height // 8,
            width // 8,
            device=device,
            dtype=torch.half if amp else torch.float,
        )

    def split_features(self, features, context_dim, feature_dim):
        context, features = torch.split(features, [context_dim, feature_dim], dim=1)

        context, _ = context.chunk(chunks=2, dim=0)
        feature0, feature1 = features.chunk(chunks=2, dim=0)

        return features, torch.relu(context)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[0.0, 0.0, 0.0],
            bgr_mult=[1.0, 1.0, 1.0],
            bgr_to_rgb=False,
            resize_mode="interpolation",
            interpolation_align_corners=False,
        )

        img0 = images[:, 0]
        img1 = images[:, 1]

        flow_list = []

        if not self.has_init_bhwd:
            self.init_bhwd(
                img0.shape[0],
                img0.shape[2],
                img0.shape[3],
                img0.device,
                img0.dtype == torch.float16,
            )

        features_s16, features_s8 = self.backbone(torch.cat([img0, img1], dim=0))

        features_s16 = self.cross_attn_s16(features_s16)

        features_s16, context_s16 = self.split_features(
            features_s16, self.context_dim_s16, self.feature_dim_s16
        )
        features_s8, context_s8 = self.split_features(
            features_s8, self.context_dim_s8, self.feature_dim_s8
        )

        feature0_s16, feature1_s16 = features_s16.chunk(chunks=2, dim=0)

        flow0 = self.matching_s16.global_correlation_softmax(feature0_s16, feature1_s16)

        # flow0 = self.flow_attn_s16(feature0_s16, flow0)

        corr_pyr_s16 = self.corr_block_s16.init_corr_pyr(feature0_s16, feature1_s16)

        iter_context_s16 = self.init_iter_context_s16

        for i in range(self.iters_s16):
            if self.training and i > 0:
                flow0 = flow0.detach()
                # iter_context_s16 = iter_context_s16.detach()

            corrs = self.corr_block_s16(corr_pyr_s16, flow0)

            iter_context_s16, delta_flow = self.refine_s16(
                corrs, context_s16, iter_context_s16, flow0
            )

            flow0 = flow0 + delta_flow

            if self.training:
                up_flow0 = F.interpolate(flow0, scale_factor=16, mode="bilinear") * 16
                up_flow0 = self.postprocess_predictions(
                    up_flow0, image_resizer, is_flow=True
                )
                flow_list.append(up_flow0)

        flow0 = F.interpolate(flow0, scale_factor=2, mode="nearest") * 2

        features_s16 = F.interpolate(features_s16, scale_factor=2, mode="nearest")

        features_s8 = self.merge_s8(torch.cat([features_s8, features_s16], dim=1))

        feature0_s8, feature1_s8 = features_s8.chunk(chunks=2, dim=0)

        corr_pyr_s8 = self.corr_block_s8.init_corr_pyr(feature0_s8, feature1_s8)

        context_s16 = F.interpolate(context_s16, scale_factor=2, mode="nearest")

        context_s8 = self.context_merge_s8(torch.cat([context_s8, context_s16], dim=1))

        iter_context_s8 = self.init_iter_context_s8

        for i in range(self.iters_s8):
            if self.training and i > 0:
                flow0 = flow0.detach()
                # iter_context_s8 = iter_context_s8.detach()

            corrs = self.corr_block_s8(corr_pyr_s8, flow0)

            iter_context_s8, delta_flow = self.refine_s8(
                corrs, context_s8, iter_context_s8, flow0
            )

            flow0 = flow0 + delta_flow

            if self.training or i == self.iters_s8 - 1:
                feature0_s1 = self.conv_s8(img0)
                up_flow0 = self.upsample_s8(feature0_s1, flow0) * 8
                up_flow0 = self.postprocess_predictions(
                    up_flow0, image_resizer, is_flow=True
                )
                flow_list.append(up_flow0)

        if self.training:
            outputs = {"flows": up_flow0[:, None], "flow_preds": flow_list}
        else:
            outputs = {"flows": up_flow0[:, None]}

        return outputs


@register_model
@trainable
class neuflow2(NeuFlow2):
    pass



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from functools import partial
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .norm import GroupNorm2d, LayerNorm2d, BatchNorm2d
from .local_timm.gelu import GELU


def compute_pyramid_levels(x):
    img_diag = math.sqrt((x.shape[-2] ** 2) + (x.shape[-1] ** 2))
    input_factor = max(
        1, img_diag / 1100
    )  # 1100 ~= math.sqrt((960 ** 2) + (540 ** 2)), i.e., 1K resolution
    pyr_levels = int(round(math.log2(input_factor))) + 3
    return pyr_levels


def bilinear_sampler(img, coords, mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.float()

    return img


def get_activation(name):
    if name == "relu":
        return nn.ReLU
    elif name == "gelu":
        return GELU
    elif name == "silu":
        return nn.SiLU
    elif name == "mish":
        return nn.Mish
    elif name == "linear":
        return nn.Identity
    else:
        return None


def get_norm(name, affine=False, num_groups=8):
    if name == "group":
        return partial(GroupNorm2d, affine=affine, num_groups=num_groups)
    elif name == "layer":
        return partial(LayerNorm2d, affine=affine)
    elif name == "batch":
        return BatchNorm2d
    else:
        return None



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import torch.nn as nn
from .conv import Conv2dBlock


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_layer, stride=1):
        super(ResidualBlock, self).__init__()

        self.conv1 = Conv2dBlock(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = Conv2dBlock(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        self.norm1 = norm_layer(num_channels=planes)
        self.norm2 = norm_layer(num_channels=planes)
        if not stride == 1:
            self.norm3 = norm_layer(num_channels=planes)

        if stride == 1 and in_planes == planes:
            self.downsample = None
        else:
            self.downsample = nn.Sequential(
                Conv2dBlock(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class ResStem(nn.Module):
    def __init__(self, hidden_chs, norm_layer):
        super(ResStem, self).__init__()
        self.norm_fn = norm_layer

        self.norm1 = norm_layer(num_channels=hidden_chs[0])

        self.conv1 = Conv2dBlock(3, hidden_chs[0], kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = hidden_chs[0]
        self.layer1 = self._make_layer(hidden_chs[0], stride=1)
        self.layer2 = self._make_layer(hidden_chs[1], stride=2)

        self.conv2 = Conv2dBlock(hidden_chs[1], hidden_chs[2], kernel_size=1)

        # for m in self.modules():
        #     if isinstance(m, nn.Conv2d):
        #         nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
        #     elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
        #         if m.weight is not None:
        #             nn.init.constant_(m.weight, 1)
        #         if m.bias is not None:
        #             nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)

        x = self.conv2(x)

        return x



import torch
import torch.nn.functional as F
import numpy as np
from scipy import interpolate


class InputPadder:
    """Pads images such that dimensions are divisible by 8"""

    def __init__(self, dims, mode="sintel", factor=8):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // 8) + 1) * 8 - self.ht) % 8
        pad_wd = (((self.wd // 8) + 1) * 8 - self.wd) % 8
        if mode == "sintel":
            self._pad = [
                pad_wd // 2,
                pad_wd - pad_wd // 2,
                pad_ht // 2,
                pad_ht - pad_ht // 2,
            ]
        else:
            self._pad = [pad_wd // 2, pad_wd - pad_wd // 2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode="replicate") for x in inputs]

    def pad_list(self, inputs):
        return [F.pad(x, self._pad, mode="replicate") for x in inputs]

    def unpad(self, x):
        # assert x.equal(torch.load('/share/sunshangkun/x.pt'))
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        # print(c)
        # assert x[..., c[0]:c[1], c[2]:c[3]].equal(torch.load('/share/sunshangkun/unpad_flows_0.pt').cuda())

        return x[..., c[0] : c[1], c[2] : c[3]]


def forward_interpolate(flow):
    flow = flow.detach().cpu().numpy()
    dx, dy = flow[0], flow[1]

    ht, wd = dx.shape
    x0, y0 = np.meshgrid(np.arange(wd), np.arange(ht))

    x1 = x0 + dx
    y1 = y0 + dy

    x1 = x1.reshape(-1)
    y1 = y1.reshape(-1)
    dx = dx.reshape(-1)
    dy = dy.reshape(-1)

    valid = (x1 > 0) & (x1 < wd) & (y1 > 0) & (y1 < ht)
    x1 = x1[valid]
    y1 = y1[valid]
    dx = dx[valid]
    dy = dy[valid]

    flow_x = interpolate.griddata(
        (x1, y1), dx, (x0, y0), method="nearest", fill_value=0
    )

    flow_y = interpolate.griddata(
        (x1, y1), dy, (x0, y0), method="nearest", fill_value=0
    )

    flow = np.stack([flow_x, flow_y], axis=0)
    return torch.from_numpy(flow).float()


def bilinear_sampler(img, coords, mode="bilinear", mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.float()

    return img


def coords_grid(batch, ht, wd, dtype, device):
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
    )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].expand(batch, -1, -1, -1)


def coords_grid_y_first(batch, ht, wd, dtype, device):
    """Place y grid before x grid"""
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
    )
    coords = torch.stack(coords, dim=0).int()
    return coords[None].expand(batch, -1, -1, -1)


def soft_argmax(corr_me, B, H1, W1):
    # Implement soft argmin
    coords, feats = corr_me.decomposed_coordinates_and_features

    # Computing soft argmin
    flow_pred = torch.zeros(B, 2, H1, W1).to(corr_me.device)
    for batch, (coord, feat) in enumerate(zip(coords, feats)):
        coord_img_1 = coord[:, :2].to(corr_me.device)
        coord_img_2 = coord[:, 2:].to(corr_me.device)
        # relative positions (flow hypotheses)
        rel_pos = coord_img_2 - coord_img_1
        # augmented indices
        aug_coord_img_1 = (coord_img_1[:, 0:1] * W1 + coord_img_1[:, 1:2]).long()
        # run softmax on the score
        weight = scatter_softmax(feat, aug_coord_img_1, dim=0)
        rel_pos_weighted = weight * rel_pos
        out = scatter_add(rel_pos_weighted, aug_coord_img_1, dim=0)
        # Need to permute (y, x) to (x, y) for flow
        flow_pred[batch] = out[:, [1, 0]].view(H1, W1, 2).permute(2, 0, 1)
    return flow_pred


def upflow8(flow, mode="bilinear"):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return 8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def upflow4(flow, mode="bilinear"):
    new_size = (4 * flow.shape[2], 4 * flow.shape[3])
    return 4 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def upflow2(flow, mode="bilinear"):
    new_size = (2 * flow.shape[2], 2 * flow.shape[3])
    return 2 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def downflow8(flow, mode="bilinear"):
    new_size = (flow.shape[2] // 8, flow.shape[3] // 8)
    return F.interpolate(flow, size=new_size, mode=mode, align_corners=True) / 8


def downflow4(flow, mode="bilinear"):
    new_size = (flow.shape[2] // 4, flow.shape[3] // 4)
    return F.interpolate(flow, size=new_size, mode=mode, align_corners=True) / 4


def compute_interpolation_weights(yx_warped):
    # yx_warped: [N, 2]
    y_warped = yx_warped[:, 0]
    x_warped = yx_warped[:, 1]

    # elementwise operations below
    y_f = torch.floor(y_warped)
    y_c = y_f + 1
    x_f = torch.floor(x_warped)
    x_c = x_f + 1

    w0 = (y_c - y_warped) * (x_c - x_warped)
    w1 = (y_warped - y_f) * (x_c - x_warped)
    w2 = (y_c - y_warped) * (x_warped - x_f)
    w3 = (y_warped - y_f) * (x_warped - x_f)

    weights = [w0, w1, w2, w3]
    indices = [
        torch.stack([y_f, x_f], dim=1),
        torch.stack([y_c, x_f], dim=1),
        torch.stack([y_f, x_c], dim=1),
        torch.stack([y_c, x_c], dim=1),
    ]
    weights = torch.cat(weights, dim=1)
    indices = torch.cat(indices, dim=2)
    # indices = torch.cat(indices, dim=0)  # [4*N, 2]

    return weights, indices


def compute_inverse_interpolation_img(weights, indices, img, b, h_i, w_i):
    """
    weights: [b, h*w]
    indices: [b, h*w]
    img: [b, h*w, a, b, c, ...]
    """
    w0, w1, w2, w3 = weights
    ff_idx, cf_idx, fc_idx, cc_idx = indices

    k = len(img.size()) - len(w0.size())
    img_0 = w0[(...,) + (None,) * k] * img
    img_1 = w1[(...,) + (None,) * k] * img
    img_2 = w2[(...,) + (None,) * k] * img
    img_3 = w3[(...,) + (None,) * k] * img

    img_out = torch.zeros(b, h_i * w_i, *img.shape[2:]).type_as(img)

    ff_idx = torch.clamp(ff_idx, min=0, max=h_i * w_i - 1)
    cf_idx = torch.clamp(cf_idx, min=0, max=h_i * w_i - 1)
    fc_idx = torch.clamp(fc_idx, min=0, max=h_i * w_i - 1)
    cc_idx = torch.clamp(cc_idx, min=0, max=h_i * w_i - 1)

    img_out.scatter_add_(1, ff_idx[(...,) + (None,) * k].expand_as(img_0), img_0)
    img_out.scatter_add_(1, cf_idx[(...,) + (None,) * k].expand_as(img_1), img_1)
    img_out.scatter_add_(1, fc_idx[(...,) + (None,) * k].expand_as(img_2), img_2)
    img_out.scatter_add_(1, cc_idx[(...,) + (None,) * k].expand_as(img_3), img_3)

    return img_out  # [b, h_i*w_i, ...]



import torch.nn.functional as F

from . import utils


class Matching:
    def init_bhwd(self, batch_size, height, width, device, amp):
        self.grid = utils.coords_grid(
            batch_size, height, width, device, amp
        )  # [B, 2, H, W]
        self.flatten_grid = self.grid.view(batch_size, 2, -1).permute(
            0, 2, 1
        )  # [B, H*W, 2]

    def global_correlation_softmax(self, feature0, feature1):
        b, c, h, w = feature0.shape

        feature0 = feature0.flatten(-2).permute(0, 2, 1)
        feature1 = feature1.flatten(-2).permute(0, 2, 1)

        correspondence = F.scaled_dot_product_attention(
            feature0, feature1, self.flatten_grid
        )

        correspondence = correspondence.view(b, h, w, 2).permute(
            0, 3, 1, 2
        )  # [B, 2, H, W]

        flow = correspondence - self.grid

        return flow



#!/usr/bin/env python

import os

import torch

import cupy
import re

kernel_Softsplat_updateOutput = """
	extern "C" __global__ void kernel_Softsplat_updateOutput(
		const int n,
		const float* input,
		const float* flow,
		double* output
	) { for (int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x; intIndex < n; intIndex += blockDim.x * gridDim.x) {
		const int intN = ( intIndex / SIZE_3(output) / SIZE_2(output) / SIZE_1(output) ) % SIZE_0(output);
		const int intC = ( intIndex / SIZE_3(output) / SIZE_2(output)                  ) % SIZE_1(output);
		const int intY = ( intIndex / SIZE_3(output)                                   ) % SIZE_2(output);
		const int intX = ( intIndex                                                    ) % SIZE_3(output);

		float fltOutputX = (float) (intX) + VALUE_4(flow, intN, 0, intY, intX);
		float fltOutputY = (float) (intY) + VALUE_4(flow, intN, 1, intY, intX);

		int intNorthwestX = (int) (floor(fltOutputX));
		int intNorthwestY = (int) (floor(fltOutputY));
		int intNortheastX = intNorthwestX + 1;
		int intNortheastY = intNorthwestY;
		int intSouthwestX = intNorthwestX;
		int intSouthwestY = intNorthwestY + 1;
		int intSoutheastX = intNorthwestX + 1;
		int intSoutheastY = intNorthwestY + 1;

		float fltNorthwest = ((float) (intSoutheastX) - fltOutputX) * ((float) (intSoutheastY) - fltOutputY);
		float fltNortheast = (fltOutputX - (float) (intSouthwestX)) * ((float) (intSouthwestY) - fltOutputY);
		float fltSouthwest = ((float) (intNortheastX) - fltOutputX) * (fltOutputY - (float) (intNortheastY));
		float fltSoutheast = (fltOutputX - (float) (intNorthwestX)) * (fltOutputY - (float) (intNorthwestY));

		float value = VALUE_4(input, intN, intC, intY, intX);
		double valueNorthwest = value * fltNorthwest;
		double valueNortheast = value * fltNortheast;
		double valueSouthwest = value * fltSouthwest;
		double valueSoutheast = value * fltSoutheast;		
		
		if ((intNorthwestX >= 0) & (intNorthwestX < SIZE_3(output)) & (intNorthwestY >= 0) & (intNorthwestY < SIZE_2(output))) {
			atomicAdd(&output[OFFSET_4(output, intN, intC, intNorthwestY, intNorthwestX)], valueNorthwest);
		}
		
		if ((intNortheastX >= 0) & (intNortheastX < SIZE_3(output)) & (intNortheastY >= 0) & (intNortheastY < SIZE_2(output))) {
			atomicAdd(&output[OFFSET_4(output, intN, intC, intNortheastY, intNortheastX)], valueNortheast);
		}
		
		if ((intSouthwestX >= 0) & (intSouthwestX < SIZE_3(output)) & (intSouthwestY >= 0) & (intSouthwestY < SIZE_2(output))) {
			atomicAdd(&output[OFFSET_4(output, intN, intC, intSouthwestY, intSouthwestX)], valueSouthwest);
		}
		
		if ((intSoutheastX >= 0) & (intSoutheastX < SIZE_3(output)) & (intSoutheastY >= 0) & (intSoutheastY < SIZE_2(output))) {
			atomicAdd(&output[OFFSET_4(output, intN, intC, intSoutheastY, intSoutheastX)], valueSoutheast);
		}
	}}
"""

kernel_Softsplat_updateGradInput = """
	extern "C" __global__ void kernel_Softsplat_updateGradInput(
		const int n,
		const float* input,
		const float* flow,
		const float* gradOutput,
		float* gradInput,
		float* gradFlow
	) { for (int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x; intIndex < n; intIndex += blockDim.x * gridDim.x) {
		const int intN = ( intIndex / SIZE_3(gradInput) / SIZE_2(gradInput) / SIZE_1(gradInput) ) % SIZE_0(gradInput);
		const int intC = ( intIndex / SIZE_3(gradInput) / SIZE_2(gradInput)                     ) % SIZE_1(gradInput);
		const int intY = ( intIndex / SIZE_3(gradInput)                                         ) % SIZE_2(gradInput);
		const int intX = ( intIndex                                                             ) % SIZE_3(gradInput);

		float fltGradInput = 0.0;

		float fltOutputX = (float) (intX) + VALUE_4(flow, intN, 0, intY, intX);
		float fltOutputY = (float) (intY) + VALUE_4(flow, intN, 1, intY, intX);

		int intNorthwestX = (int) (floor(fltOutputX));
		int intNorthwestY = (int) (floor(fltOutputY));
		int intNortheastX = intNorthwestX + 1;
		int intNortheastY = intNorthwestY;
		int intSouthwestX = intNorthwestX;
		int intSouthwestY = intNorthwestY + 1;
		int intSoutheastX = intNorthwestX + 1;
		int intSoutheastY = intNorthwestY + 1;

		float fltNorthwest = ((float) (intSoutheastX) - fltOutputX) * ((float) (intSoutheastY) - fltOutputY);
		float fltNortheast = (fltOutputX - (float) (intSouthwestX)) * ((float) (intSouthwestY) - fltOutputY);
		float fltSouthwest = ((float) (intNortheastX) - fltOutputX) * (fltOutputY - (float) (intNortheastY));
		float fltSoutheast = (fltOutputX - (float) (intNorthwestX)) * (fltOutputY - (float) (intNorthwestY));

		if ((intNorthwestX >= 0) & (intNorthwestX < SIZE_3(gradOutput)) & (intNorthwestY >= 0) & (intNorthwestY < SIZE_2(gradOutput))) {
			fltGradInput += VALUE_4(gradOutput, intN, intC, intNorthwestY, intNorthwestX) * fltNorthwest;
		}

		if ((intNortheastX >= 0) & (intNortheastX < SIZE_3(gradOutput)) & (intNortheastY >= 0) & (intNortheastY < SIZE_2(gradOutput))) {
			fltGradInput += VALUE_4(gradOutput, intN, intC, intNortheastY, intNortheastX) * fltNortheast;
		}

		if ((intSouthwestX >= 0) & (intSouthwestX < SIZE_3(gradOutput)) & (intSouthwestY >= 0) & (intSouthwestY < SIZE_2(gradOutput))) {
			fltGradInput += VALUE_4(gradOutput, intN, intC, intSouthwestY, intSouthwestX) * fltSouthwest;
		}

		if ((intSoutheastX >= 0) & (intSoutheastX < SIZE_3(gradOutput)) & (intSoutheastY >= 0) & (intSoutheastY < SIZE_2(gradOutput))) {
			fltGradInput += VALUE_4(gradOutput, intN, intC, intSoutheastY, intSoutheastX) * fltSoutheast;
		}

		gradInput[intIndex] = fltGradInput;
	} }
"""

kernel_Softsplat_updateGradFlow = """
	extern "C" __global__ void kernel_Softsplat_updateGradFlow(
		const int n,
		const float* input,
		const float* flow,
		const float* gradOutput,
		float* gradInput,
		float* gradFlow
	) { for (int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x; intIndex < n; intIndex += blockDim.x * gridDim.x) {
		float fltGradFlow = 0.0;

		const int intN = ( intIndex / SIZE_3(gradFlow) / SIZE_2(gradFlow) / SIZE_1(gradFlow) ) % SIZE_0(gradFlow);
		const int intC = ( intIndex / SIZE_3(gradFlow) / SIZE_2(gradFlow)                    ) % SIZE_1(gradFlow);
		const int intY = ( intIndex / SIZE_3(gradFlow)                                       ) % SIZE_2(gradFlow);
		const int intX = ( intIndex                                                          ) % SIZE_3(gradFlow);

		float fltOutputX = (float) (intX) + VALUE_4(flow, intN, 0, intY, intX);
		float fltOutputY = (float) (intY) + VALUE_4(flow, intN, 1, intY, intX);

		int intNorthwestX = (int) (floor(fltOutputX));
		int intNorthwestY = (int) (floor(fltOutputY));
		int intNortheastX = intNorthwestX + 1;
		int intNortheastY = intNorthwestY;
		int intSouthwestX = intNorthwestX;
		int intSouthwestY = intNorthwestY + 1;
		int intSoutheastX = intNorthwestX + 1;
		int intSoutheastY = intNorthwestY + 1;

		float fltNorthwest = 0.0;
		float fltNortheast = 0.0;
		float fltSouthwest = 0.0;
		float fltSoutheast = 0.0;

		if (intC == 0) {
			fltNorthwest = ((float) (-1.0)) * ((float) (intSoutheastY) - fltOutputY);
			fltNortheast = ((float) (+1.0)) * ((float) (intSouthwestY) - fltOutputY);
			fltSouthwest = ((float) (-1.0)) * (fltOutputY - (float) (intNortheastY));
			fltSoutheast = ((float) (+1.0)) * (fltOutputY - (float) (intNorthwestY));

		} else if (intC == 1) {
			fltNorthwest = ((float) (intSoutheastX) - fltOutputX) * ((float) (-1.0));
			fltNortheast = (fltOutputX - (float) (intSouthwestX)) * ((float) (-1.0));
			fltSouthwest = ((float) (intNortheastX) - fltOutputX) * ((float) (+1.0));
			fltSoutheast = (fltOutputX - (float) (intNorthwestX)) * ((float) (+1.0));

		}

		for (int intChannel = 0; intChannel < SIZE_1(gradOutput); intChannel += 1) {
			float fltInput = VALUE_4(input, intN, intChannel, intY, intX);

			if ((intNorthwestX >= 0) & (intNorthwestX < SIZE_3(gradOutput)) & (intNorthwestY >= 0) & (intNorthwestY < SIZE_2(gradOutput))) {
				fltGradFlow += fltInput * VALUE_4(gradOutput, intN, intChannel, intNorthwestY, intNorthwestX) * fltNorthwest;
			}

			if ((intNortheastX >= 0) & (intNortheastX < SIZE_3(gradOutput)) & (intNortheastY >= 0) & (intNortheastY < SIZE_2(gradOutput))) {
				fltGradFlow += fltInput * VALUE_4(gradOutput, intN, intChannel, intNortheastY, intNortheastX) * fltNortheast;
			}

			if ((intSouthwestX >= 0) & (intSouthwestX < SIZE_3(gradOutput)) & (intSouthwestY >= 0) & (intSouthwestY < SIZE_2(gradOutput))) {
				fltGradFlow += fltInput * VALUE_4(gradOutput, intN, intChannel, intSouthwestY, intSouthwestX) * fltSouthwest;
			}

			if ((intSoutheastX >= 0) & (intSoutheastX < SIZE_3(gradOutput)) & (intSoutheastY >= 0) & (intSoutheastY < SIZE_2(gradOutput))) {
				fltGradFlow += fltInput * VALUE_4(gradOutput, intN, intChannel, intSoutheastY, intSoutheastX) * fltSoutheast;
			}
		}

		gradFlow[intIndex] = fltGradFlow;
	} }
"""


def cupy_kernel(strFunction, objVariables):
    strKernel = globals()[strFunction]

    while True:
        objMatch = re.search("(SIZE_)([0-4])(\()([^\)]*)(\))", strKernel)

        if objMatch is None:
            break
        # end

        intArg = int(objMatch.group(2))

        strTensor = objMatch.group(4)
        intSizes = objVariables[strTensor].size()

        strKernel = strKernel.replace(objMatch.group(), str(intSizes[intArg]))
    # end

    while True:
        objMatch = re.search("(OFFSET_)([0-4])(\()([^\)]+)(\))", strKernel)

        if objMatch is None:
            break
        # end

        intArgs = int(objMatch.group(2))
        strArgs = objMatch.group(4).split(",")

        strTensor = strArgs[0]
        intStrides = objVariables[strTensor].stride()
        strIndex = [
            "(("
            + strArgs[intArg + 1].replace("{", "(").replace("}", ")").strip()
            + ")*"
            + str(intStrides[intArg])
            + ")"
            for intArg in range(intArgs)
        ]

        strKernel = strKernel.replace(
            objMatch.group(0), "(" + str.join("+", strIndex) + ")"
        )
    # end

    while True:
        objMatch = re.search("(VALUE_)([0-4])(\()([^\)]+)(\))", strKernel)

        if objMatch is None:
            break
        # end

        intArgs = int(objMatch.group(2))
        strArgs = objMatch.group(4).split(",")

        strTensor = strArgs[0]
        intStrides = objVariables[strTensor].stride()
        strIndex = [
            "(("
            + strArgs[intArg + 1].replace("{", "(").replace("}", ")").strip()
            + ")*"
            + str(intStrides[intArg])
            + ")"
            for intArg in range(intArgs)
        ]

        strKernel = strKernel.replace(
            objMatch.group(0), strTensor + "[" + str.join("+", strIndex) + "]"
        )
    # end

    return strKernel


# end


@cupy.memoize(for_each_device=True)
def cupy_launch(strFunction, strKernel):
    # return cupy.RawModule(code=strKernel, options=tuple(['-I ' + os.environ['CUDA_HOME'], '-I ' + os.environ['CUDA_HOME'] + '/include'])).get_function(strFunction)
    return cupy.RawModule(code=strKernel).get_function(strFunction)


# end


class _FunctionSoftsplat(torch.autograd.Function):
    @staticmethod
    def forward(self, input, flow):
        intSamples = input.shape[0]
        intInputDepth, intInputHeight, intInputWidth = (
            input.shape[1],
            input.shape[2],
            input.shape[3],
        )
        intFlowDepth, intFlowHeight, intFlowWidth = (
            flow.shape[1],
            flow.shape[2],
            flow.shape[3],
        )

        assert intFlowDepth == 2
        assert intInputHeight == intFlowHeight
        assert intInputWidth == intFlowWidth

        input = input.contiguous()
        assert input.is_cuda == True
        flow = flow.contiguous()
        assert flow.is_cuda == True

        output = input.new_zeros(
            [intSamples, intInputDepth, intInputHeight, intInputWidth]
        ).double()

        if input.is_cuda == True:
            n = output.nelement()
            cupy_launch(
                "kernel_Softsplat_updateOutput",
                cupy_kernel(
                    "kernel_Softsplat_updateOutput",
                    {"input": input, "flow": flow, "output": output},
                ),
            )(
                grid=tuple([int((n + 512 - 1) / 512), 1, 1]),
                block=tuple([512, 1, 1]),
                args=[
                    cupy.int32(n),
                    input.data_ptr(),
                    flow.data_ptr(),
                    output.data_ptr(),
                ],
            )

        elif input.is_cuda == False:
            raise NotImplementedError()

        # end

        self.save_for_backward(input, flow)

        return output.float()

    # end

    @staticmethod
    def backward(self, gradOutput):
        input, flow = self.saved_tensors

        intSamples = input.shape[0]
        intInputDepth, intInputHeight, intInputWidth = (
            input.shape[1],
            input.shape[2],
            input.shape[3],
        )
        intFlowDepth, intFlowHeight, intFlowWidth = (
            flow.shape[1],
            flow.shape[2],
            flow.shape[3],
        )

        assert intFlowDepth == 2
        assert intInputHeight == intFlowHeight
        assert intInputWidth == intFlowWidth

        gradOutput = gradOutput.contiguous()
        assert gradOutput.is_cuda == True

        gradInput = (
            input.new_zeros([intSamples, intInputDepth, intInputHeight, intInputWidth])
            if self.needs_input_grad[0] == True
            else None
        )
        gradFlow = (
            input.new_zeros([intSamples, intFlowDepth, intFlowHeight, intFlowWidth])
            if self.needs_input_grad[1] == True
            else None
        )

        if input.is_cuda == True:
            if gradInput is not None:
                n = gradInput.nelement()
                cupy_launch(
                    "kernel_Softsplat_updateGradInput",
                    cupy_kernel(
                        "kernel_Softsplat_updateGradInput",
                        {
                            "input": input,
                            "flow": flow,
                            "gradOutput": gradOutput,
                            "gradInput": gradInput,
                            "gradFlow": gradFlow,
                        },
                    ),
                )(
                    grid=tuple([int((n + 512 - 1) / 512), 1, 1]),
                    block=tuple([512, 1, 1]),
                    args=[
                        cupy.int32(n),
                        input.data_ptr(),
                        flow.data_ptr(),
                        gradOutput.data_ptr(),
                        gradInput.data_ptr(),
                        None,
                    ],
                )
            # end

            if gradFlow is not None:
                n = gradFlow.nelement()
                cupy_launch(
                    "kernel_Softsplat_updateGradFlow",
                    cupy_kernel(
                        "kernel_Softsplat_updateGradFlow",
                        {
                            "input": input,
                            "flow": flow,
                            "gradOutput": gradOutput,
                            "gradInput": gradInput,
                            "gradFlow": gradFlow,
                        },
                    ),
                )(
                    grid=tuple([int((n + 512 - 1) / 512), 1, 1]),
                    block=tuple([512, 1, 1]),
                    args=[
                        cupy.int32(n),
                        input.data_ptr(),
                        flow.data_ptr(),
                        gradOutput.data_ptr(),
                        None,
                        gradFlow.data_ptr(),
                    ],
                )
            # end

        elif input.is_cuda == False:
            raise NotImplementedError()

        # end

        return gradInput, gradFlow

    # end


# end


def FunctionSoftsplat(tenInput, tenFlow, tenMetric=None, strType="average"):
    assert tenMetric is None or tenMetric.shape[1] == 1
    assert strType in ["summation", "average", "linear", "softmax"]

    if strType == "average":
        tenInput = torch.cat(
            [
                tenInput,
                tenInput.new_ones(
                    tenInput.shape[0], 1, tenInput.shape[2], tenInput.shape[3]
                ),
            ],
            1,
        )

    elif strType == "linear":
        tenInput = torch.cat([tenInput * tenMetric, tenMetric], 1)

    elif strType == "softmax":
        tenInput = torch.cat([tenInput * tenMetric.exp(), tenMetric.exp()], 1)

    # end

    tenOutput = _FunctionSoftsplat.apply(tenInput, tenFlow)

    if strType != "summation":
        tenNormalize = tenOutput[:, -1:, :, :]

        tenNormalize[tenNormalize == 0.0] = 1.0

        tenOutput = tenOutput[:, :-1, :, :] / tenNormalize
    # end

    return tenOutput


# end


class ModuleSoftsplat(torch.nn.Module):
    def __init__(self, strType):
        super().__init__()

        self.strType = strType

    # end

    def forward(self, tenInput, tenFlow, tenMetric):
        return FunctionSoftsplat(tenInput, tenFlow, tenMetric, self.strType)

    # end


# end
